from .snapshot import Snapshot
from .generic_repr import GenericRepr
from .module import assert_match_snapshot
from .unittest import TestCase


__all__ = ['Snapshot', 'GenericRepr', 'assert_match_snapshot', 'TestCase']



from __future__ import absolute_import
from django.test import TestCase as dTestCase
from django.test.runner import DiscoverRunner

from snapshottest.reporting import reporting_lines
from .unittest import TestCase as uTestCase


class TestRunner(DiscoverRunner):

    separator1 = "=" * 70
    separator2 = "-" * 70

    def __init__(self, snapshot_update=False, **kwargs):
        super(TestRunner, self).__init__(**kwargs)
        TestCase.snapshot_should_update = snapshot_update

    @classmethod
    def add_arguments(cls, parser):
        super(TestRunner, cls).add_arguments(parser)
        parser.add_argument(
            '--snapshot-update', default=False, action='store_true',
            dest='snapshot_update', help='Update the snapshots automatically.',
        )

    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        result = super(TestRunner, self).run_tests(
                test_labels=test_labels,
                extra_tests=extra_tests,
                **kwargs
        )
        self.print_report()
        return result

    def print_report(self):
        print("\n" + self.separator1)
        print('SnapshotTest summary')
        print(self.separator2)
        for line in reporting_lines('python manage.py test'):
            print(line)
        print(self.separator1)


class TestCase(uTestCase, dTestCase):
    pass



from difflib import Differ
from termcolor import colored

from .sorted_dict import SortedDict
from .formatter import Formatter


def format_line(line):
    line = line.rstrip('\n')
    if line.startswith('-'):
        return colored(line, 'green', attrs=['bold'])
    elif line.startswith('+'):
        return colored(line, 'red', attrs=['bold'])
    elif line.startswith('?'):
        return (
            colored('') +
            colored(line, 'yellow', attrs=['bold'])
        )

    return colored('') + colored(line, 'white', attrs=['dark'])


class PrettyDiff(object):
    def __init__(self, obj, snapshottest):
        if isinstance(obj, dict):
            obj = SortedDict(**obj)
        self.obj = obj
        self.pretty = Formatter()
        self.differ = Differ()
        self.snapshottest = snapshottest

    def __eq__(self, other):
        return isinstance(other, PrettyDiff) and self.obj == other.obj

    def __repr__(self):
        return repr(self.obj)

    def get_diff(self, other):
        text1 = ['Received ', ''] + self.pretty(self.obj).splitlines(1)
        text2 = ['Snapshot ', ''] + self.pretty(other).splitlines(1)

        lines = list(self.differ.compare(text2, text1))
        return [
            format_line(line) for line in lines
        ]



from __future__ import absolute_import
import logging
import os

from nose.plugins import Plugin

from .module import SnapshotModule
from .reporting import reporting_lines
from .unittest import TestCase

log = logging.getLogger('nose.plugins.snapshottest')


class SnapshotTestPlugin(Plugin):
    name = 'snapshottest'
    enabled = True

    separator1 = "=" * 70
    separator2 = "-" * 70

    def options(self, parser, env=os.environ):
        super(SnapshotTestPlugin, self).options(parser, env=env)
        parser.add_option(
            '--snapshot-update',
            action='store_true',
            default=False,
            dest='snapshot_update',
            help='Update the snapshots.'
        )
        parser.add_option(
            '--snapshot-disable',
            action='store_true',
            dest='snapshot_disable',
            default=False,
            help="Disable special SnapshotTest"
        )

    def configure(self, options, conf):
        super(SnapshotTestPlugin, self).configure(options, conf)
        self.snapshot_update = options.snapshot_update
        self.enabled = not options.snapshot_disable

    def wantClass(self, cls):
        if issubclass(cls, TestCase):
            cls._snapshot_should_update = self.snapshot_update

    def afterContext(self):
        if self.snapshot_update:
            for module in SnapshotModule.get_modules():
                module.delete_unvisited()
                module.save()

    def report(self, stream):
        if not SnapshotModule.has_snapshots():
            return

        stream.writeln(self.separator1)
        stream.writeln('SnapshotTest summary')
        stream.writeln(self.separator2)
        for line in reporting_lines('nosetests'):
            stream.writeln(line)
        stream.writeln(self.separator1)



class GenericRepr(object):
    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        representation = repr(self.obj)
        # We remove the hex id, if found
        representation = representation.replace(hex(id(self.obj)), "0x100000000")
        return 'GenericRepr("{}")'.format(representation)

    def __eq__(self, other):
        return isinstance(other, GenericRepr) and repr(self) == repr(other)



import six

from .sorted_dict import SortedDict
from .generic_repr import GenericRepr


def trepr(s):
    text = '\n'.join([repr(line)[1:-1] for line in s.split('\n')])
    quotes, dquotes = "'''", '"""'
    if quotes in text:
        if dquotes in text:
            text = text.replace(quotes, "\\'\\'\\'")
        else:
            quotes = dquotes
    return "%s%s%s" % (quotes, text, quotes)


class Formatter(object):
    def __init__(self, imports=None):
        self.types = {}
        self.htchar = ' '*4
        self.lfchar = '\n'
        self.indent = 0
        self.imports = imports

    def set_formater(self, obj, callback):
        self.types[obj] = callback

    def __call__(self, value, **args):
        return self.format(value, self.indent)

    def format(self, value, indent):
        from .diff import PrettyDiff
        if value is None:
            return 'None'
        if isinstance(value, PrettyDiff):
            return self.format(value.obj, indent)
        if isinstance(value, dict):
            return self.format_dict(value, indent)
        elif isinstance(value, tuple):
            return self.format_tuple(value, indent)
        elif isinstance(value, list):
            return self.format_list(value, indent)
        elif isinstance(value, six.string_types):
            return self.format_str(value, indent)
        elif isinstance(value, (int, float, complex, bool, bytes, set, frozenset, GenericRepr)):
            return self.format_std_type(value, indent)

        return self.format_object(value, indent)

    def format_str(self, value, indent):
        if '\n' in value:
            # Is a multiline string, so we use '''{}''' for the repr
            return trepr(value)

        return repr(str(value))

    def format_std_type(self, value, indent):
        return repr(value)

    def format_object(self, value, indent):
        if self.imports:
            self.imports['snapshottest'].add('GenericRepr')
        return repr(GenericRepr(value))

    def format_dict(self, value, indent):
        value = SortedDict(**value)
        items = [
            self.lfchar + self.htchar * (indent + 1) + self.format(key, indent) + ': ' +
            self.format(value[key], indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + self.lfchar + self.htchar * indent)

    def format_list(self, value, indent):
        items = [
            self.lfchar + self.htchar * (indent + 1) + self.format(item, indent + 1)
            for item in value
        ]
        return '[%s]' % (','.join(items) + self.lfchar + self.htchar * indent)

    def format_tuple(self, value, indent):
        items = [
            self.lfchar + self.htchar * (indent + 1) + self.format(item, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + self.lfchar + self.htchar * indent)



import os
from termcolor import colored

from .module import SnapshotModule


def reporting_lines(testing_cli):
    successful_snapshots = SnapshotModule.stats_successful_snapshots()
    bold = ['bold']
    if successful_snapshots:
        yield (
            colored('{} snapshots passed', attrs=bold) + '.'
        ).format(successful_snapshots)
    new_snapshots = SnapshotModule.stats_new_snapshots()
    if new_snapshots[0]:
        yield (
            colored('{} snapshots written', 'green', attrs=bold) + ' in {} test suites.'
        ).format(*new_snapshots)
    inspect_str = colored(
        'Inspect your code or run with `{} --snapshot-update` to update them.'.format(testing_cli),
        attrs=['dark']
    )
    failed_snapshots = SnapshotModule.stats_failed_snapshots()
    if failed_snapshots[0]:
        yield (
            colored('{} snapshots failed', 'red', attrs=bold) + ' in {} test suites. '
            + inspect_str
        ).format(*failed_snapshots)
    unvisited_snapshots = SnapshotModule.stats_unvisited_snapshots()
    if unvisited_snapshots[0]:
        yield (
            colored('{} snapshots deprecated', 'yellow', attrs=bold) + ' in {} test suites. '
            + inspect_str
        ).format(*unvisited_snapshots)


def diff_report(left, right):
    return [
        'stored snapshot should match the received value',
        '',
        colored('> ') +
        colored('Received value', 'red', attrs=['bold']) +
        colored(' does not match ', attrs=['bold']) +
        colored('stored snapshot `{}`'.format(
            left.snapshottest.test_name,
        ), 'green', attrs=['bold']) +
        colored('.', attrs=['bold']),
        colored('') + '> ' + os.path.relpath(left.snapshottest.module.filepath, os.getcwd()),
        '',
    ] + left.get_diff(right)



import os
import imp
from collections import defaultdict
import logging

from .snapshot import Snapshot
from .formatter import Formatter
from .diff import PrettyDiff
# from .error import SnapshotError


logger = logging.getLogger(__name__)


class SnapshotModule(object):
    _snapshot_modules = {}

    def __init__(self, module, filepath):
        self._original_snapshot = None
        self._snapshots = None
        self.module = module
        self.filepath = filepath
        self.imports = defaultdict(set)
        self.visited_snapshots = set()
        self.new_snapshots = set()
        self.failed_snapshots = set()
        self.imports['snapshottest'].add('Snapshot')

    def load_snapshots(self):
        try:
            source = imp.load_source(self.module, self.filepath)
            assert isinstance(source.snapshots, Snapshot)
            return source.snapshots
        except:
            return Snapshot()

    def visit(self, snapshot_name):
        self.visited_snapshots.add(snapshot_name)

    def delete_unvisited(self):
        for unvisited in self.unvisited_snapshots:
            del self.snapshots[unvisited]

    @property
    def unvisited_snapshots(self):
        return set(self.snapshots.keys()) - self.visited_snapshots

    @classmethod
    def total_unvisited_snapshots(cls):
        unvisited_snapshots = 0
        unvisited_modules = 0
        for module in cls.get_modules():
            l = len(module.unvisited_snapshots)
            unvisited_snapshots += l
            unvisited_modules += min(l, 1)

        return unvisited_snapshots, unvisited_modules

    @classmethod
    def get_modules(cls):
        return SnapshotModule._snapshot_modules.values()

    @classmethod
    def stats_for_module(cls, getter):
        count_snapshots = 0
        count_modules = 0
        for module in SnapshotModule._snapshot_modules.values():
            l = getter(module)
            count_snapshots += l
            count_modules += min(l, 1)

        return count_snapshots, count_modules

    @classmethod
    def stats_unvisited_snapshots(cls):
        return cls.stats_for_module(lambda module: len(module.unvisited_snapshots))

    @classmethod
    def stats_visited_snapshots(cls):
        return cls.stats_for_module(lambda module: len(module.visited_snapshots))

    @classmethod
    def stats_new_snapshots(cls):
        return cls.stats_for_module(lambda module: len(module.new_snapshots))

    @classmethod
    def stats_failed_snapshots(cls):
        return cls.stats_for_module(lambda module: len(module.failed_snapshots))

    @classmethod
    def stats_successful_snapshots(cls):
        stats_visited = cls.stats_visited_snapshots()
        stats_failed = cls.stats_failed_snapshots()
        return stats_visited[0] - stats_failed[0]

    @classmethod
    def has_snapshots(cls):
        return cls.stats_visited_snapshots()[0] > 0

    @property
    def original_snapshot(self):
        if not self._original_snapshot:
            self._original_snapshot = self.load_snapshots()
        return self._original_snapshot

    @property
    def snapshots(self):
        if not self._snapshots:
            self._snapshots = Snapshot(self.original_snapshot)
        return self._snapshots

    def __getitem__(self, key):
        return self.snapshots.get(key, None)

    def __setitem__(self, key, value):
        if key not in self.snapshots:
            # It's a new test
            self.new_snapshots.add(key)
        self.snapshots[key] = value

    def mark_failed(self, key):
        return self.failed_snapshots.add(key)

    def save(self):
        if self.original_snapshot == self.snapshots:
            # If there are no changes, we do nothing
            return

        snapshot_dir = os.path.dirname(self.filepath)

        # Create the snapshot dir in case doesn't exist
        try:
            os.makedirs(snapshot_dir, 0o0700)
        except (IOError, OSError):
            pass

        # Create __init__.py in case doesn't exist
        open(os.path.join(snapshot_dir, '__init__.py'), 'a').close()

        pretty = Formatter(self.imports)

        with open(self.filepath, 'w') as snapshot_file:
            snapshots_declarations = []
            for key, value in self.snapshots.items():
                snapshots_declarations.append('''snapshots['{}'] = {}'''.format(key, pretty(value)))

            imports = '\n'.join([
                'from {} import {}'.format(module, ', '.join(module_imports))
                for module, module_imports in self.imports.items()
            ])
            snapshot_file.write('''# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

{}


snapshots = Snapshot()

{}
'''.format(imports, '\n\n'.join(snapshots_declarations)))

    @classmethod
    def get_module_for_testpath(cls, test_filepath):
        if test_filepath not in cls._snapshot_modules:
            dirname = os.path.dirname(test_filepath)
            snapshot_dir = os.path.join(dirname, "snapshots")

            snapshot_basename = 'snap_{}.py'.format(os.path.splitext(os.path.basename(test_filepath))[0])
            snapshot_filename = os.path.join(snapshot_dir, snapshot_basename)
            snapshot_module = '{}'.format(os.path.splitext(snapshot_basename)[0])

            cls._snapshot_modules[test_filepath] = SnapshotModule(snapshot_module, snapshot_filename)

        return cls._snapshot_modules[test_filepath]


class SnapshotTest(object):
    _current_tester = None
    update = False

    def __init__(self):
        self.curr_snapshot = 1

    @property
    def module(self):
        raise NotImplementedError("module property needs to be implemented")

    @property
    def update(self):
        return False

    @property
    def test_name(self):
        raise NotImplementedError("test_name property needs to be implemented")

    def __enter__(self):
        SnapshotTest._current_tester = self
        return self

    def __exit__(self, type, value, tb):
        self.save_changes()
        SnapshotTest._current_tester = None

    def visit(self):
        self.module.visit(self.test_name)

    def fail(self):
        self.module.mark_failed(self.test_name)

    def store(self, data):
        self.module[self.test_name] = data
        self.curr_snapshot += 1

    def assert_equals(self, value, snapshot):
        assert value == snapshot

    def assert_match(self, value):
        self.visit()
        prev_snapshot = not self.update and self.module[self.test_name]
        if prev_snapshot:
            try:
                self.assert_equals(
                    PrettyDiff(value, self),
                    PrettyDiff(prev_snapshot, self)
                )
            except:
                self.fail()
                raise

        self.store(value)

    def save_changes(self):
        self.module.save()


def assert_match_snapshot(value):
    if not SnapshotTest._current_tester:
        raise Exception("You need to use assert_match_snapshot in the SnapshotTest context.")

    SnapshotTest._current_tester.assert_match(value)



from collections import OrderedDict


class SortedDict(OrderedDict):
    def __init__(self, **kwargs):
        super(SortedDict, self).__init__()

        for key, value in sorted(kwargs.items()):
            if isinstance(value, dict):
                self[key] = SortedDict(**value)
            elif isinstance(value, list):
                self[key] = self._sort_list(value)
            else:
                self[key] = value

    def _sort_list(self, value):
        def sort(val):
            if isinstance(val, dict):
                return SortedDict(**val)
            elif isinstance(val, list):
                return self._sort_list(val)
            else:
                return val

        return [sort(item) for item in value]



from __future__ import absolute_import
import unittest
import inspect

from .module import SnapshotModule, SnapshotTest
from .diff import PrettyDiff
from .reporting import diff_report


class UnitTestSnapshotTest(SnapshotTest):

    def __init__(self, test_class, test_id, test_filepath, should_update, assertEqual):
        self.test_class = test_class
        self.test_id = test_id
        self.test_filepath = test_filepath
        self.assertEqual = assertEqual
        self.should_update = should_update
        super(UnitTestSnapshotTest, self).__init__()

    @property
    def module(self):
        return SnapshotModule.get_module_for_testpath(self.test_filepath)

    @property
    def update(self):
        return self.should_update

    def assert_equals(self, value, snapshot):
        self.assertEqual(value, snapshot)

    @property
    def test_name(self):
        class_name = self.test_class.__name__
        test_name = self.test_id.split('.')[-1]
        return '{}::{} {}'.format(
            class_name,
            test_name,
            self.curr_snapshot
        )


# Inspired by https://gist.github.com/twolfson/13f5f5784f67fd49b245
class TestCase(unittest.TestCase):

    snapshot_should_update = False

    @classmethod
    def setUpClass(cls):
        """On inherited classes, run our `setUp` method"""
        cls._snapshot_tests = []
        cls._snapshot_file = inspect.getfile(cls)

        if cls is not TestCase and cls.setUp is not TestCase.setUp:
            orig_setUp = cls.setUp
            orig_tearDown = cls.tearDown

            def setUpOverride(self, *args, **kwargs):
                TestCase.setUp(self)
                return orig_setUp(self, *args, **kwargs)

            def tearDownOverride(self, *args, **kwargs):
                TestCase.tearDown(self)
                return orig_tearDown(self, *args, **kwargs)

            cls.setUp = setUpOverride
            cls.tearDown = tearDownOverride

        super(TestCase, cls).setUpClass()

    def comparePrettyDifs(self, obj1, obj2, msg):
        # self
        # assert obj1 == obj2
        if not(obj1 == obj2):
            raise self.failureException('\n'.join(diff_report(obj1, obj2)))
        #     raise self.failureException("DIFF")

    @classmethod
    def tearDownClass(cls):
        if cls._snapshot_tests:
            module = SnapshotModule.get_module_for_testpath(cls._snapshot_file)
            module.save()
        super(TestCase, cls).tearDownClass()

    def setUp(self):
        """Do some custom setup"""
        # print dir(self.__module__)
        self.addTypeEqualityFunc(PrettyDiff, self.comparePrettyDifs)
        self._snapshot = UnitTestSnapshotTest(
            test_class=self.__class__,
            test_id=self.id(),
            test_filepath=self._snapshot_file,
            should_update=self.snapshot_should_update,
            assertEqual=self.assertEqual
        )
        self._snapshot_tests.append(self._snapshot)
        SnapshotTest._current_tester = self._snapshot

    def tearDown(self):
        """Do some custom setup"""
        # print dir(self.__module__)
        SnapshotTest._current_tester = None
        self._snapshot = None

    def assert_match_snapshot(self, value):
        self._snapshot.assert_match(value)

    assertMatchSnapshot = assert_match_snapshot



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

tests_require = ['six', 'pytest>=3.0', 'pytest-cov', 'nose', 'django>=1.10.6']

setup(
    name='snapshottest',
    version='0.5.0',
    description='Snapshot Testing utils for Python',
    long_description=readme,
    author='Syrus Akbary',
    author_email='me@syrusakbary.com',
    url='https://github.com/syrusakbary/snapshottest',
    # custom PyPI classifier for pytest plugins
    entry_points={
        'pytest11': [
            'snapshottest = snapshottest.pytest',
        ],
        'nose.plugins.0.10':
        ['snapshottest = snapshottest.nose:SnapshotTestPlugin']
    },
    install_requires=['six>=1.10.0', 'termcolor'],
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
        'pytest': [
            'pytest',
        ],
        'nose': [
            'nose',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: OS Independent',
        'Framework :: Pytest',
    ],
    license='MIT',
    packages=find_packages(exclude=('tests', )))



from __future__ import absolute_import
import pytest

from .module import SnapshotModule, SnapshotTest
from .diff import PrettyDiff
from .reporting import reporting_lines, diff_report


def pytest_addoption(parser):
    group = parser.getgroup('snapshottest')
    group.addoption(
        '--snapshot-update',
        action='store_true',
        default=False,
        dest='snapshot_update',
        help='Update the snapshots.'
    )
    group.addoption(
        '--snapshot-verbose',
        action='store_true',
        default=False,
        help='Dump diagnostic and progress information.'
    )


class PyTestSnapshotTest(SnapshotTest):

    def __init__(self, request=None):
        self.request = request
        super(PyTestSnapshotTest, self).__init__()

    @property
    def module(self):
        return SnapshotModule.get_module_for_testpath(self.request.node.fspath.strpath)

    @property
    def update(self):
        return self.request.config.option.snapshot_update

    @property
    def test_name(self):
        cls_name = getattr(self.request.node.cls, '__name__', '')
        return '{}{} {}'.format(
            '{}.'.format(cls_name) if cls_name else '',
            self.request.node.name,
            self.curr_snapshot
        )


class SnapshotSession(object):
    def __init__(self, config):
        self.verbose = config.getoption("snapshot_verbose")
        self.config = config

    def display(self, tr):
        if not SnapshotModule.has_snapshots():
            return

        tr.write_sep("=", "SnapshotTest summary")

        for line in reporting_lines('pytest'):
            tr.write_line(line)


def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, PrettyDiff) and op == "==":
        return diff_report(left, right)


@pytest.fixture
def snapshot(request):
    with PyTestSnapshotTest(request) as snapshot_test:
        yield snapshot_test


def pytest_terminal_summary(terminalreporter):
    if terminalreporter.config.option.snapshot_update:
        for module in SnapshotModule.get_modules():
            module.delete_unvisited()
            module.save()

    terminalreporter.config._snapshotsession.display(terminalreporter)


@pytest.mark.trylast  # force the other plugins to initialise, fixes issue with capture not being properly initialised
def pytest_configure(config):
    config._snapshotsession = SnapshotSession(config)
    # config.pluginmanager.register(bs, "snapshottest")
